import numpy as np
import PIL.Image as Image
import PIL.ImageColor as ImageColor
import PIL.ImageDraw as ImageDraw
import PIL.ImageFont as ImageFont


STANDARD_COLORS = list(ImageColor.colormap)


def draw_bounding_box_from_array(image, box, scores=None, labels=None, colors="red", font_color="black", 
                                 thickness=2, font_size=12, use_normalized_coordinates=True):
    """
    Adds a bounding box to an image (numpy array).
    
    Arguments:
        + image (np.array) -- Image to draw the bounding box with shape [height, width, 3].
        + box (list) -- Coordinates from the box. [xmin, ymin, xmax, ymax].
        + scores (list, np.array) -- Output scores from the FERModel.
        + labels (list, dict) -- Labels from each score.
        + colors (str, dict) -- Colors to draw bounding box. Default is 'red'. 
            A dict can be passed to draw boxes of different colors for each label.
        + font_color -- Color to draw the text.
        + thickness (int) -- Line thickness. Default value is 4.
        + font_size (int) -- Font size to draw the text.
        + use_normalized_coordinates (bool) -- If True (default), treat coordinates
            [xmin, ymin, xmax, ymax] as relative to the image.  Otherwise treat
            coordinates as absolute.
    """

    pil_image = Image.fromarray(np.uint8(image)).convert("RGB")
    draw_bounding_box(pil_image, box, scores, labels, colors, font_color, 
                      thickness, font_size, use_normalized_coordinates)
    np.copyto(image, np.array(pil_image))


def draw_bounding_box(image, box, scores=None, labels=None, colors="red", font_color="black", 
                      thickness=2, font_size=12, use_normalized_coordinates=True):
    """
    Adds a bounding box to an image.
    
    Arguments:
        + image (PIL.Image) -- A PIL.Image object.
        + box (list) -- Coordinates from the box. [xmin, ymin, xmax, ymax].
        + scores (list, np.array) -- Output scores from the FERModel.
        + labels (list, dict) -- Labels from each score.
        + colors (str, dict) -- Colors to draw bounding box. Default is 'red'. 
            A dict can be passed to draw boxes of different colors for each label.
        + font_color -- Color to draw the text.
        + thickness (int) -- Line thickness. Default value is 4.
        + font_size (int) -- Font size to draw the text.
        + use_normalized_coordinates (bool) -- If True (default), treat coordinates
            [xmin, ymin, xmax, ymax] as relative to the image.  Otherwise treat
            coordinates as absolute.
    """

    if isinstance(colors, dict):
        assert scores is not None, "If scores is None, colors must be a color name (str) rather than a dictionary."
    else:
        color = colors

    # Init drawing tool
    draw = ImageDraw.Draw(image)

    # labels list to dict
    if isinstance(labels, list):
        labels = {i:l for i, l in enumerate(labels)} 

    # Retrieve info from scores
    if scores is not None:
        # most likely label
        max_score = np.argmax(scores)
        # Select color and text based on scores
        prob = scores[max_score]
        if labels is not None:
            text = labels[max_score] + ": {:.2f}%".format(prob*100)
        else:
            text = str(max_score) + ": {:.2f}%".format(prob*100)
        color = colors if not isinstance(colors, dict) else colors[max_score]

    else:
        text = None

    # Unpack image shape and box coordinates
    w, h = image.size
    # ymin, xmin, ymax, xmax = box
    xmin, ymin, xmax, ymax = box

    # Get box coordinates
    if use_normalized_coordinates:
        (left, right, top, bottom) = (xmin * w, xmax * w, ymin * h, ymax * h)
    else:
        (left, right, top, bottom) = (xmin, xmax, ymin, ymax)

    # Draw box
    draw.line([(left, top), (left, bottom), (right, bottom),
               (right, top), (left, top)], width=thickness, fill=color)

    # Write label
    if text is not None:
        draw_text(image, text, left, right, top, bottom, color, font_color, font_size)


def draw_bounding_boxes_from_array(image, boxes, scores=None, labels=None, colors="red", font_color="black",
                                   thickness=2, font_size=12, use_normalized_coordinates=True):
    """
    Adds multiples bounding boxes to an image (numpy array).
    
    Arguments:
        + image (np.array) -- Image to draw the bounding box with shape [height, width, 3].
        + boxes (list) -- Coordinates from the boxes. [[xmin, ymin, xmax, ymax], ...].
        + scores (list, np.array) -- Output scores from the FERModel.
        + labels (list, dict) -- Labels from each score.
        + colors (str, dict) -- Colors to draw bounding box. Default is 'red'. 
            A dict can be passed to draw boxes of different colors for each label.
        + font_color -- Color to draw the text.
        + thickness (int) -- Line thickness. Default value is 4.
        + font_size (int) -- Font size to draw the text.
        + use_normalized_coordinates (bool) -- If True (default), treat coordinates
            [[xmin, ymin, xmax, ymax], ...] as relative to the image.  Otherwise treat
            coordinates as absolute.
    """

    pil_image = Image.fromarray(np.uint8(image)).convert("RGB")
    draw_bounding_boxes(pil_image, boxes, scores, labels, colors, font_color, 
                        thickness, font_size, use_normalized_coordinates)
    np.copyto(image, np.array(pil_image))


def draw_bounding_boxes(image, boxes, scores=None, labels=None, colors="red", font_color="black", 
                        thickness=2, font_size=12, use_normalized_coordinates=True):
    """
    Adds multiples bounding boxes to an image.
    
    Arguments:
        + image (PIL.Image) -- A PIL.Image object.
        + boxes (list) -- Coordinates from the boxes. [[xmin, ymin, xmax, ymax], ...].
        + scores (list, np.array) -- Output scores from the FERModel.
        + labels (list, dict) -- Labels from each score.
        + colors (str, dict) -- Colors to draw bounding box. Default is 'red'. 
            A dict can be passed to draw boxes of different colors for each label.
        + font_color -- Color to draw the text.
        + thickness (int) -- Line thickness. Default value is 4.
        + font_size (int) -- Font size to draw the text.
        + use_normalized_coordinates (bool) -- If True (default), treat coordinates
            [[xmin, ymin, xmax, ymax], ...] as relative to the image.  Otherwise treat
            coordinates as absolute.
    """

    assert np.array(
        boxes).shape[1] == 4, "boxes must be a list of shape (N, 4)."

    if scores is not None:
        assert len(
            boxes) == scores.shape[0], "The number of boxes and scores must be the same."

    for i, box in enumerate(boxes):
        score = None if scores is None else scores[i]
        draw_bounding_box(image, box, score, labels, colors, font_color, thickness, 
                          font_size, use_normalized_coordinates)


def draw_text_from_array(image, text, left, right, top, bottom, color="red", font_color="black", font_size=12):
    """
    Adds a text box to an image (numpy array).
    
    Arguments:
        + image (np.array) -- Image to draw a text with shape [height, width, 3]
        + text (str) -- Text to draw on the image.
        + left (int) -- left point in the rectangle containg the text.
        + right (int) -- left point in the rectangle containg the text.
        + top (int) -- left point in the rectangle containg the text.
        + bottom (int) -- left point in the rectangle containg the text.
        + color (str) -- Color to draw the text box. Default is 'red'. 
        + font_color -- Color to draw the text.
        + font_size (int) -- Font size to draw the text.
    """

    pil_image = Image.fromarray(np.uint8(image)).convert("RGB")
    draw_text(pil_image, text, left, right, top, bottom, color, font_color, font_size)
    np.copyto(image, np.array(pil_image))


def draw_text(image, text, left, right, top, bottom, color="red", font_color="black", font_size=12):
    """
    Adds a text box to an image (numpy array).
    
    Arguments:
        + image (PIL.Image) --  A PIL.Image object.
        + text (str) -- Text to draw on the image.
        + left (int) -- left point in the rectangle containg the text.
        + right (int) -- left point in the rectangle containg the text.
        + top (int) -- left point in the rectangle containg the text.
        + bottom (int) -- left point in the rectangle containg the text.
        + color (str) -- Color to draw the text box. Default is 'red'. 
        + font_color -- Color to draw the text.
        + font_size (int) -- Font size to draw the text.
    """

    # Init drawing tool
    draw = ImageDraw.Draw(image)
    
    # Load font
    try:
        font = ImageFont.truetype('arial.ttf', font_size)
    except IOError:
        font = ImageFont.load_default()
        
    # If the total height of the display strings added to the top of the bounding
    # box exceeds the top of the image, stack the strings below the bounding box
    # instead of above.
    text_width, text_height = font.getsize(text)
    # Each display_str has a top and bottom margin of 0.05x.
    total_text_height = (1 + 2 * 0.05) *  text_height
    text_bottom = top if top > total_text_height else bottom + total_text_height
    margin = np.ceil(0.05 * text_height)
    
    # Draw rectangle
    x0, x1 = left, left + text_width
    y0, y1 = text_bottom - text_height - 2 * margin, text_bottom
    draw.rectangle([(x0, y0), (x1, y1)], fill=color)
    
    # Draw text
    draw.text((left + margin, text_bottom - text_height - margin), 
              text, fill=font_color, font=font)


import os
import json
import warnings
from tensorflow.keras.utils import get_file

from . import vggface
# from . import inception (to be implemented)


FILENAME = os.path.join(os.getenv("HOME"), ".keras/models/deep_expressions/models.json")


def _get_models_json_file():
    """Download models.json file from DeepExpressions models repo."""

    get_file("models.json", 
             "https://raw.githubusercontent.com/deepexpressions/models/master/models.json",
             cache_subdir = "models/deep_expressions/")


def _model_exists(name):
    """Check if a model exists in DeepExpressions models."""
    
    # Remove layer identity from name
    name = name.strip('_l3|_l1|+')

    # Download models.json file
    _get_models_json_file()

    # Load models.json
    with open(FILENAME, "r") as f:
        data = json.load(f)

    return True if name in data.keys() else False

    
def _get_model_labels(name):
    """Get labels given a model name."""

    # Remove layer identity from name
    name = name.strip('_l3|_l1|+')

    # Check if model_name exists
    if not _model_exists(name):
        raise Exception("`{}` is not a valid name for a model.".format(name))
    
    # Load models.json
    with open(FILENAME, "r") as f:
        data = json.load(f)

    # Return model labels
    labels = data[name]["labels"]
    return {int(k) : v for k, v in labels.items()}


def print_model_info(name):
    """Prints DeepExpressions model informations."""

    # Remove layer identity from name
    name = name.strip('_l3|_l1|+')

    # Check if model_name exists
    if not _model_exists(name):
        raise Exception("`{}` is not a valid name for a model.".format(name))
    
    # Load models.json
    with open(FILENAME, "r") as f:
        data = json.load(f)


    # Print model info
    print("DeepExpressions models:")
    print("-----------------------")
    print(" * ", name)
    print("    ConvNet: ", data[name]["convnet"])
    print("    Database: ", data[name]["database"])
    print("    Derivations: ", data[name]["derivations"])
    print("    Labels: ", list(data[name]["labels"].values()), "\n")


def list_models(only_names=True):
    """List DeepExpressions trained models informations."""

    # Download models.json file
    _get_models_json_file()
    
    # Load models.json
    with open(FILENAME, "r") as f:
        data = json.load(f)

    # Print models name
    if only_names:
        print("DeepExpressions models:")
        print("-----------------------")
        [print(" * ", name) for name in data]

    # Print models info
    else:
        print("DeepExpressions models:")
        print("-----------------------")

        for name in data:
            print(" * ", name)
            print("    ConvNet: ", data[name]["convnet"])
            print("    Database: ", data[name]["database"])
            print("    Derivations: ", data[name]["derivations"])
            print("    Labels: ", list(data[name]["labels"].values()), "\n")


def load_model(name, input_shape):
    """
    Load a Facial Expression Recognition classifier from DeepExpressions models.

    Arguments:
        + name (str) -- Name of the model. Options available at : ?.

    Output:
        + Keras model for Facial Expression Recognition.
        + Preprocess input function
    """

    print("[INFO] Loading model '{}' ...".format(name), end=" ")

    if name in {"vggface_ce", "vggface_ce_l1"}:
        model = vggface.VGGFACE(weights="ce", num_layers=1 , num_classes=7, input_shape=input_shape)

    elif name == "vggface_ce_l3":
        model = vggface.VGGFACE(weights="ce", num_layers=3 , num_classes=7, input_shape=input_shape)
    
    elif name in {"vggface_ck", "vggface_ck+", "vggface_ck_l1", "vggface_ck+_l1"}:
        model = vggface.VGGFACE(weights="ck", num_layers=1 , num_classes=7, input_shape=input_shape)

    elif name in {"vggface_ck_l3", "vggface_ck+_l3"}:
        model = vggface.VGGFACE(weights="ck", num_layers=3 , num_classes=7, input_shape=input_shape)

    elif name in {"vggface_kdef", "vggface_kdef_l1"}:
        model = vggface.VGGFACE(weights="kdef", num_layers=1 , num_classes=7, input_shape=input_shape)

    elif name == "vggface_kdef_l3":
        model = vggface.VGGFACE(weights="kdef", num_layers=3 , num_classes=7, input_shape=input_shape)

    elif name == "vggface_iffg":
        model = vggface.VGGFACE(weights="iffg", num_layers=3, num_classes=9, input_shape=input_shape)

    else:
        raise Exception("Unknown model.")

    print("Done.")

    return model


def load_model_from_dir(filename):
    """
    Load a Facial Expression Recognition classifier from DeepExpressions models.

    Arguments:
        + filename (str) -- Path to custom model saved (only *.h5 files).

    Output:
        + Keras model for Facial Expression Recognition.
    """

    # Verify file extension
    assert filename.endswith(".h5"), "Custom model must be save with extenstion '.h5'."

    print("[INFO] Loading model '{}' ...".format(name), end=" ")
    model = tf.keras.models.load_model(filename)
    print("Done.")

    return model    


def _obtain_input_shape(input_shape,
                        default_size,
                        min_size,
                        data_format,
                        require_flatten,
                        weights=None):
    """Internal utility to compute/validate a model's input shape.
    # Arguments
        input_shape: Either None (will return the default network input shape),
            or a user-provided shape to be validated.
        default_size: Default input width/height for the model.
        min_size: Minimum input width/height accepted by the model.
        data_format: Image data format to use.
        require_flatten: Whether the model is expected to
            be linked to a classifier via a Flatten layer.
        weights: One of `None` (random initialization)
            or 'imagenet' (pre-training on ImageNet).
            If weights='imagenet' input channels must be equal to 3.
    # Returns
        An integer shape tuple (may include None entries).
    # Raises
        ValueError: In case of invalid argument values.
    """
    if weights != 'imagenet' and input_shape and len(input_shape) == 3:
        if data_format == 'channels_first':
            if input_shape[0] not in {1, 3}:
                warnings.warn(
                    'This model usually expects 1 or 3 input channels. '
                    'However, it was passed an input_shape with ' +
                    str(input_shape[0]) + ' input channels.')
            default_shape = (input_shape[0], default_size, default_size)
        else:
            if input_shape[-1] not in {1, 3}:
                warnings.warn(
                    'This model usually expects 1 or 3 input channels. '
                    'However, it was passed an input_shape with ' +
                    str(input_shape[-1]) + ' input channels.')
            default_shape = (default_size, default_size, input_shape[-1])
    else:
        if data_format == 'channels_first':
            default_shape = (3, default_size, default_size)
        else:
            default_shape = (default_size, default_size, 3)
    if weights == 'imagenet' and require_flatten:
        if input_shape is not None:
            if input_shape != default_shape:
                raise ValueError('When setting `include_top=True` '
                                 'and loading `imagenet` weights, '
                                 '`input_shape` should be ' +
                                 str(default_shape) + '.')
        return default_shape
    if input_shape:
        if data_format == 'channels_first':
            if input_shape is not None:
                if len(input_shape) != 3:
                    raise ValueError(
                        '`input_shape` must be a tuple of three integers.')
                if input_shape[0] != 3 and weights == 'imagenet':
                    raise ValueError('The input must have 3 channels; got '
                                     '`input_shape=' + str(input_shape) + '`')
                if ((input_shape[1] is not None and input_shape[1] < min_size) or
                   (input_shape[2] is not None and input_shape[2] < min_size)):
                    raise ValueError('Input size must be at least ' +
                                     str(min_size) + 'x' + str(min_size) +
                                     '; got `input_shape=' +
                                     str(input_shape) + '`')
        else:
            if input_shape is not None:
                if len(input_shape) != 3:
                    raise ValueError(
                        '`input_shape` must be a tuple of three integers.')
                if input_shape[-1] != 3 and weights == 'imagenet':
                    raise ValueError('The input must have 3 channels; got '
                                     '`input_shape=' + str(input_shape) + '`')
                if ((input_shape[0] is not None and input_shape[0] < min_size) or
                   (input_shape[1] is not None and input_shape[1] < min_size)):
                    raise ValueError('Input size must be at least ' +
                                     str(min_size) + 'x' + str(min_size) +
                                     '; got `input_shape=' +
                                     str(input_shape) + '`')
    else:
        if require_flatten:
            input_shape = default_shape
        else:
            if data_format == 'channels_first':
                input_shape = (3, None, None)
            else:
                input_shape = (None, None, 3)
    if require_flatten:
        if None in input_shape:
            raise ValueError('If `include_top` is True, '
                             'you should specify a static `input_shape`. '
                             'Got `input_shape=' + str(input_shape) + '`')
    return input_shape


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import pathlib
import tensorflow as tf
from functools import partial
from . import data_augmentation as daug


# Autotune
AUTOTUNE = tf.data.experimental.AUTOTUNE

# Norm P [0., 1.]
_norm_p = lambda image, label: (image / 255.0, label)
# Norm P [-1., 1.]
_norm_n = lambda image, label: ((image / 127.5) - 1, label)
# Clip P [0., 1.]
_clip_p = lambda image, label: (tf.clip_by_value(image, 0.0, 1.0), label)
# Clip P [-1., 1.]
_clip_n = lambda image, label: (tf.clip_by_value(image, -1.0, 1.0), label)


class DataLoader():
    """
    DataLoader is an instance to load image data from a folder using tensorflow>=2.0.0-alpha0.

    Arguments:
        + data_root (str) -- Path to dataset folder.
        + image_size (int) -- Size to resize the loaded images.
        + data_root (str) -- Directory containing sub-folder for each label.
        + batch_size (int) -- Dataset batch size.
        + steps_per_epoch (int) -- Number of steps per epoch.
        + cache (bool) -- Load dataset in cache.
        + normalize (str) -- If "norm_b": normalize image to range [0, 1].
                             If "norm_n": normalize image to range [-1, 1].
                             Else do not normalize.
        + flip_left_right (bool) -- Data augmentation: Flip an image horizontally (left to right).
        + flip_up_down (bool) -- Data augmentation: Flip an image vertically (upside down).
        + random_brightness (float) -- Data augmentation: Adjust the brightness of images by a random factor.
        + random_contrast (tuple) -- Data augmentation: Adjust the contrast of an image or images by a random factor.
        + random_hue (float) -- Data augmentation: Adjust the hue of RGB images by a random factor.
        + random_jpeg_quality (tuple) -- Data augmentation: Randomly changes jpeg encoding quality for inducing jpeg noise.
        + random_rotation (float) -- Data augmentation: Performs a random rotation of a image.
        + random_saturation (tuple) -- Data augmentation: Adjust the saturation of RGB images by a random factor.
        + random_shear (float) -- Data augmentation: Performs a random spatial shear of a image.
        + random_zoom (tuple) -- Data augmentation: Performs a random spatial zoom of a image.

    Methods:
        + show_sample_batch -- show data from the first batch.

    Attributes:
        + ds -- dataset in batches.
        + image_count -- number of images from dataset.
        + label_names -- list with labels.
        + label_to_index -- dict with indexed labels.
    """


    def __init__(self, data_root, image_size=128, batch_size=32, 
        steps_per_epoch=None, cache=False, normalize=None, 
        flip_left_right=False, flip_up_down=False,
        random_brightness=None, random_contrast=None,
        random_hue=None, random_jpeg_quality=None,
        random_rotation=None, random_saturation=None,
        random_shear=None, random_zoom=None):

        # Pass input variables
        self.batch_size = batch_size
        self.image_size = image_size

        # Load all images and their labels
        all_image_paths, all_image_labels = self._load_all_images_and_labels(
            data_root)

        # Get the number of images loaded
        self.image_count = len(all_image_paths)
        _original_image_count = len(all_image_paths)

        # Pass steps_per_epoch variable
        if steps_per_epoch is None:
            self.steps_per_epoch = self._calc_steps_per_epoch()
        else:
            self.steps_per_epoch = steps_per_epoch

        # Load the data
        ds = tf.data.Dataset.from_tensor_slices(
            (all_image_paths, all_image_labels))
        ds = ds.map(
            self._load_and_preprocess_from_path_label, num_parallel_calls=AUTOTUNE)

        # Copy dataset for augment original only
        ds_augm = ds

        # Apply data augmentation
        if flip_left_right:
            ds_augm_aux = ds.map(daug.flip_left_right, num_parallel_calls=AUTOTUNE)
            ds_augm = ds_augm.concatenate(ds_augm_aux)
            self.image_count += _original_image_count

        if flip_up_down:
            ds_augm_aux = ds.map(daug.flip_up_down, num_parallel_calls=AUTOTUNE)
            ds_augm = ds_augm.concatenate(ds_augm_aux)
            self.image_count += _original_image_count

        if random_brightness is not None:
            if not (isinstance(random_brightness, int) or isinstance(random_brightness, float)):
                raise Exception("random_brightness must be and integer or float.")
            
            daug_func = partial(daug.random_brightness, max_delta=random_brightness)
            ds_augm_aux = ds.map(daug_func, num_parallel_calls=AUTOTUNE)
            ds_augm = ds_augm.concatenate(ds_augm_aux)
            self.image_count += _original_image_count

        if random_contrast is not None:
            if not (len(random_contrast) !=2 or len([False for v in random_contrast if not(
                isinstance(v, int) or isinstance(v, float))])==0):
                raise Exception("random_contrast must be a 2 values tuple of integer or floats")

            lower, upper = random_contrast
            daug_func = partial(daug.random_contrast, lower=lower, upper=upper)
            ds_augm_aux = ds.map(daug_func, num_parallel_calls=AUTOTUNE)
            ds_augm = ds_augm.concatenate(ds_augm_aux)
            self.image_count += _original_image_count

        if random_hue is not None:
            if not (isinstance(random_hue, int) or isinstance(random_hue, float)):
                raise Exception("random_hue must be and integer or float.")
            
            daug_func = partial(daug.random_hue, max_delta=random_hue)
            ds_augm_aux = ds.map(daug_func, num_parallel_calls=AUTOTUNE)
            ds_augm = ds_augm.concatenate(ds_augm_aux)
            self.image_count += _original_image_count

        if random_jpeg_quality is not None:
            if not (len(random_jpeg_quality) !=2 or len([False for v in random_jpeg_quality if not(
                isinstance(v, int) or isinstance(v, float))])==0):
                raise Exception("random_jpeg_quality must be a 2 values tuple of integer or floats")

            lower, upper = random_jpeg_quality
            daug_func = partial(daug.random_jpeg_quality, min_jpeg_quality=lower, max_jpeg_quality=upper)
            ds_augm_aux = ds.map(daug_func, num_parallel_calls=AUTOTUNE)
            ds_augm = ds_augm.concatenate(ds_augm_aux)
            self.image_count += _original_image_count

        if random_rotation is not None:
            if not (isinstance(random_rotation, int) or isinstance(random_rotation, float)):
                raise Exception("random_rotation must be and integer or float.")
            
            daug_func = partial(daug.random_rotation, angle=random_rotation)
            ds_augm_aux = ds.map(daug_func, num_parallel_calls=AUTOTUNE)
            ds_augm = ds_augm.concatenate(ds_augm_aux)
            self.image_count += _original_image_count

        if random_saturation is not None:
            if not (len(random_saturation) !=2 or len([False for v in random_saturation if not(
                isinstance(v, int) or isinstance(v, float))])==0):
                raise Exception("random_saturation must be a 2 values tuple of integer or floats")

            lower, upper = random_saturation
            daug_func = partial(daug.random_saturation, lower=lower, upper=upper)
            ds_augm_aux = ds.map(daug_func, num_parallel_calls=AUTOTUNE)
            ds_augm = ds_augm.concatenate(ds_augm_aux)
            self.image_count += _original_image_count

        if random_shear is not None:
            if not (isinstance(random_shear, int) or isinstance(random_shear, float)):
                raise Exception("random_shear must be and integer or float.")
            
            daug_func = partial(daug.random_shear, intensity=random_shear)
            ds_augm_aux = ds.map(daug_func, num_parallel_calls=AUTOTUNE)
            ds_augm = ds_augm.concatenate(ds_augm_aux)
            self.image_count += _original_image_count

        if random_zoom is not None:
            if not (len(random_zoom) !=2 or len([False for v in random_zoom if not(
                isinstance(v, int) or isinstance(v, float))])==0):
                raise Exception("random_zoom must be a 2 values tuple of integer or floats")

            daug_func = partial(daug.random_zoom, zoom_range=random_zoom)
            ds_augm_aux = ds.map(daug_func, num_parallel_calls=AUTOTUNE)
            ds_augm = ds_augm.concatenate(ds_augm_aux)
            self.image_count += _original_image_count

        # Replace origal dataset for the augmented one
        ds = ds_augm

        ## Preprocess dataset
        # By norm P [0,1]
        if normalize == "norm_p":
            ds = ds.map(_norm_p, num_parallel_calls=AUTOTUNE)
            ds = ds.map(_clip_p, num_parallel_calls=AUTOTUNE)
        # By norm N [-1,1]
        elif normalize == "norm_n":
            ds = ds.map(_norm_n, num_parallel_calls=AUTOTUNE)
            ds = ds.map(_clip_n, num_parallel_calls=AUTOTUNE)

        # Cache dataset
        if cache:
            ds = ds.cache()

        # Shuffle and batch dataset
        ds = ds.shuffle(buffer_size=self.image_count).repeat()
        self.ds = ds.batch(self.batch_size).prefetch(buffer_size=AUTOTUNE)


    def _load_all_images_and_labels(self, data_root):
        data_root = pathlib.Path(data_root)

        # List the available data paths
        all_image_paths = [str(p) for p in list(data_root.glob("*/*"))]
        # List the available labels
        self.label_names = sorted(
            item.name for item in data_root.glob("*/") if item.is_dir())
        # Assign an index to each label:
        self.label_to_index = dict((name, index)
                                   for index, name in enumerate(self.label_names))
        # Create a list of every file, and its label index
        all_image_labels = [self.label_to_index[pathlib.Path(
            path).parent.name] for path in all_image_paths]

        return all_image_paths, all_image_labels


    def _decode_and_resize_image(self, image):
        """Pre-process loaded images."""
        image = tf.image.decode_jpeg(image, channels=3)
        image = tf.image.resize(image, [self.image_size, self.image_size])
        return image


    def _load_and_preprocess_image(self, image_path):
        """Load and pre-process images."""
        image = tf.io.read_file(image_path)
        return self._decode_and_resize_image(image)


    def _load_and_preprocess_from_path_label(self, path, label):
        """Load and pre-process images and its respectives labels."""
        return self._load_and_preprocess_image(path), label


    def _calc_steps_per_epoch(self):
        """Calculate steps_per_epoch if none is given."""
        return tf.math.ceil(self.image_count/self.batch_size).numpy()


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import numpy as np
import tensorflow as tf
import PIL.Image as Image


def imread(image_path, as_array=False, normalize=None):
    """
    Read image file.

    Arguments:
        + image_path (str) -- Path to image file.
        + as_array (bool) -- Return numpy array instead of tf.Tensor.
        + normalize (str) -- If "norm_b": normalize image to range [0, 1].
                             If "norm_n": normalize image to range [-1, 1].
                             Else do not normalize.

    Output:
        + An image (as Tensor or array)
    """

    image = tf.io.read_file(image_path)
    image = tf.image.decode_jpeg(image, channels=3)

    image = image/255 if normalize == "norm_p" else ((
        tf.dtypes.cast(image, tf.float32)/127.5)-1) if normalize == "norm_n" else image

    image = image if not as_array else image.numpy()
    return image


def imresize(image, height=128, width=128, as_array=False):
    """
    Resize image.

    Arguments:
        + image (tf.Tensor) -- Image to be resized
        + height (int) -- New image height.
        + width (int) -- New image width.
        + as_array (bool) -- Return numpy array instead of tf.Tensor.

        Output:
        + A resized image (as Tensor or array)
    """
    image = tf.image.resize(image, [height, width]) 
    image = image if not as_array else image.numpy()
    return image


def imsave(image, filename):
    """
    Save image.

    Arguments:
        + image (np.ndarray, tf.Tensor) -- Image to be saved.
        + filename (str) -- Name of the file to be saved.
    """

    # Convert from tf.Tensor to np.ndarray
    if isinstance(image, tf.Tensor):
        image = image.numpy()
    # Convert from np.ndarray to PIL.Image object
    if isinstance(image, np.ndarray):
        image = Image.fromarray(np.uint8(image)).convert("RGB")
    # Save image
    image.save(filename)


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os
import numpy as np
import tensorflow as tf

from . import model_utils


class FERModel():
    """
    Facial Expression Recognition Model.

    Arguments:
        + model_name (str) -- Name of the model. Options available at : http://github.com/deepexpressions/models.
        + filename (str) -- Path to custom model saved (only *.h5 files).

    Methods:
        + predict -- Predict facial expressions from input_data.
        + summary -- Prints a summary representation of the model loaded.
        
    Attributes:
        + model_name -- name of the model.
        + model -- Keras Sequential model.
    """


    def __init__(self, name, labels=None, input_shape=(150, 150, 3)):
        """Model constructor."""

        self.name = name
        self.input_shape = input_shape

        # Load DeepExpressions trained model
        if not os.path.exists(name):
            self.model = model_utils.load_model(name, input_shape)
            self.labels = labels if labels is not None else model_utils._get_model_labels(name)
            self._tm = True # Use to know if the loaded model is from
                            # DeepExpressions or local file

        # Load local model
        else:
            if labels is None:
                raise Exception("Can't use local file as model without define its `labels`")

            self.model = model_utils.load_model_from_dir(name)
            self.labels = labels
            self._tm = False


    def predict(self, input_data, return_labels=False, return_scores=False):
        """
        Predict facial expressions from input data.

        Arguments:
            + input_data -- Images to predict. It should be preprocessed.
            + return_labels (bool) -- Returns predictions with label names.
            + return_scores (bool) -- Returns scores from each class in the prediction.

        Output:
            + (list of predictions, [labeled predictions], [scores])
        """
        
        # Predict data
        scores = self.model.predict(input_data)
        predictions = tf.argmax(scores, axis=1)

        output = (predictions, )

        if return_labels:
            labeled_predictions = list(map(self._label_predictions, predictions.numpy().tolist()))
            output += (labeled_predictions, )

        if return_scores:
            output += (scores, )

        return output


    def preprocess_input(self, input_data, normalize=None):
        """
        Preprocess input_data before predictions.

        Arguments:
            + input_data -- Images to predict. It can be np.ndarray, tf.Tensor or a list of the previous types.
            + normalize (str) -- If "norm_b": normalize image to range [0, 1].
                                 If "norm_n": normalize image to range [-1, 1].
                                 Else do not normalize.

        Output:
            + batch of preprocessed images.
        """

        h, w = self.input_shape[:2]

        # preprocess_func = {norm_p, norm_n, None}
        preprocess_func = lambda x: x/255 if normalize == "norm_p" else (
            (tf.dtypes.cast(x, tf.float32)/127.5)-1) if normalize == "norm_n" else x

        # input_data is a single image
        if isinstance(input_data, (np.ndarray, tf.Tensor)):
            input_data = tf.image.resize([input_data], size=(h, w))
            input_data = preprocess_func(input_data)

        # input_data are multiple images
        elif isinstance(input_data, list):
            images = []

            for image in input_data:
                images.append(
                    tf.image.resize(image, size=[h, w])
                )

            input_data = tf.convert_to_tensor(images, dtype=tf.float32)
            input_data = preprocess_func(input_data)

        else:
            raise "Invalid input_data"

        return input_data


    def info(self):
        """Prints DeepExpressions model informations."""

        if not self._tm:
            print("Local file loaded. No information could be found.")

        else:
            model_utils.print_model_info(self.name)


    def summary(self):
        """Prints a summary representation of the model loaded."""
        self.model.summary()


    def _label_predictions(self, x): 
        """Convert one-hot-encoded predictions to label names."""
        return self.labels[x]





from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from math import floor


def _extract_square_roi_by_min(box):
    # Calculate box shape and differences
    xmin, ymin, xmax, ymax = box
    box_height = ymax - ymin
    box_width = xmax - xmin
    diff = floor(abs(box_height - box_width) / 2)
    
    # Make the width and height even
    xmax = xmax if box_width % 2 == 0 else xmax - 1
    ymax = ymax if box_height % 2 == 0 else ymax - 1
    
    # Decreases box height
    if box_height > box_width:
        ymin += diff
        ymax -= diff
        
    # Decreases box width
    elif box_height < box_width:    
        xmin += diff
        xmax -= diff
        
    return [xmin, ymin, xmax, ymax]


def _extract_square_roi_by_max(box, height, width):
    # Calculate box shape and differences
    xmin, ymin, xmax, ymax = box    
    box_height = ymax - ymin
    box_width = xmax - xmin
    diff = floor(abs(box_height - box_width) / 2)
    
    # Make the width and height even
    xmax = xmax if box_width % 2 == 0 else xmax - 1
    ymax = ymax if box_height % 2 == 0 else ymax - 1

    # Increases box width
    if box_height > box_width:
        xmin = xmin - diff if (xmin - diff) > 0 else 0
        xmax = xmax + diff if (xmin + diff) < width else width - 1
        
    # Increases box height
    elif box_height < box_width:    
        ymin = ymin - diff if (ymin - diff) > 0 else 0
        ymax = ymax + diff if (ymax + diff) < height else height - 1
        
    # If box is not square, it means the box exceed the image bordes
    # Therefore, the box will be recalculated based on the new values
    # using the 'min' method
    if (ymax - ymin) != (xmax - xmin):
        xmin, ymin, xmax, ymax = _extract_square_roi_by_min([xmin, ymin, xmax, ymax])  

    return [xmin, ymin, xmax, ymax]


def extract_rois(image, boxes, method="min", square_box=False):
    """
    Extracts crops from image based on boxes of detection.

    Arguments:
        + image (np.ndarray, tf.Tensor) -- Image to be cropped
        + boxes (list) -- Coordinates from the boxes. [[xmin, ymin, xmax, ymax], ...].
        + method (str) -- Method used to get square boxes.
        + square_box (bool) -- Square the boxes before crop the image.

    Output:
        + list of cropped images based on input image and detections boxes
    """

    # Get image shape
    h, w = image.shape[:2]
    rois = []
    
    # Loop over boxes
    for box in boxes:
        # If box must be square
        if square_box:
            if method == "min":
                box = _extract_square_roi_by_min(box)
                
            elif method == "max":
                box = _extract_square_roi_by_max(box, h, w)
        
        # Crop image and append it to list
        xmin, ymin, xmax, ymax = box
        roi = image[ymin:ymax, xmin:xmax]
        rois.append(roi)
                
    return rois


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="deep-expressions",
    version="0.0.3",
    author="Humberto da Silva Neto",
    author_email="humberto.nt4@gmail.com",
    description="A Deep Learning toolkit for Facial Expressions Recognition (FER)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/deepexpressions/DeepExpressions",
    packages=setuptools.find_packages("src"),
    package_dir={"": "src"},
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",''
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'opencv-python>=3.4.0', 
        'Pillow>=6.0.0', 
    ]
)


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os
import cv2
from numpy import array
from .download_model import maybe_download_model


class Detector():
    """
    Face Detector based on OpenCV's dnn module. 
    It utilizes the Single Shot Detector (SSD) framework with a ResNet as the base network.

    Methods:
        + compute -- Compute bounding boxes from detected faces within an image using OpenCV dnn module.

    Attributes:
        + detector -- Caffe model to detect faces.
    """
    

    def __init__(self):
        """FaceDetector constructor."""

        # Download OpenCV model if it does not exists
        maybe_download_model()

        # Files path
        model_path = os.path.join(os.environ['HOME'], ".keras/opencv/face_detector/")
        model_caffe_path = os.path.join(model_path, "res10_300x300_ssd_iter_140000_fp16.caffemodel")
        proto_caffe_path = os.path.join(model_path, "deploy.prototxt")

        # Loading Caffe model to OpenCV's deep learning face detector
        print("[INFO] Loading face detector model...", end="", flush=True)
        self.detector = cv2.dnn.readNetFromCaffe(proto_caffe_path, model_caffe_path)
        print("Done.")


    def compute(self, image, size=(300,300), threshold=0.5):
        """
        Compute bounding boxes from detected faces within an image using OpenCV dnn module.

        Arguments:
            + image (np.ndarray) -- Image from which the faces will be detected.
            + size (tuple) -- New size to resize input image before detections.
            + threshold (float) -- Confidence level in detections.

        Output:
            + List of bounding boxes with shape [x0, y0, x1, y1].
        """

        # Construct input blob
        (h, w) = image.shape[:2]
        blob = cv2.dnn.blobFromImage(cv2.resize(image, size), 1.0, 
                                     size, (104.0, 177.0, 123.0))

        # Forward pass
        self.detector.setInput(blob)
        detections = self.detector.forward()

        # Track bounding boxes
        bb_list = []

        # Loop over detections:
        for i in range(0, detections.shape[2]):
            # Extract confidence probability
            confidence = detections[0,0,i,2]

            # Filter weak detections
            if confidence > threshold:
                box = detections[0, 0, i, 3:7] * array([w, h, w, h])
                bb_list.append(box.astype("int"))

        return bb_list





import tensorflow as tf
from functools import partial


def flip_left_right(image, label: tuple):
    """
    Data augmentation: Flip an image horizontally (left to right).
    
    Arguments:
        + image (tf.Tensor) -- Image to transform.
        + label (tf.Tensor) -- Label from the image.
    """
    return tf.image.flip_left_right(image), label


def flip_up_down(image, label: tuple):
    """
    Data augmentation: Flip an image vertically (upside down).
    
    Arguments:
        + image (tf.Tensor) -- Image to transform.
        + label (tf.Tensor) -- Label from the image.
    """
    return tf.image.flip_up_down(image), label


def random_brightness(image, label: tuple, max_delta=0.5):
    """
    Data augmentation: Adjust the brightness of images by a random factor.
    
    Arguments:
        + image (tf.Tensor) -- Image to transform.
        + label (tf.Tensor) -- Label from the image.
        + max_delta (float) -- Amount to add to the pixel values.
    """
    tf_image = tf.image.random_brightness(image, max_delta=max_delta)
    return tf_image, label


def random_contrast(image, label: tuple, lower=0.5, upper=2.5):
    """
    Data augmentation: Adjust the contrast of an image or images by a random factor.
    
    Arguments:
        + image (tf.Tensor) -- Image to transform.
        + label (tf.Tensor) -- Label from the image.
        + lower (float) -- Lower bound for the random contrast factor.
        + upper (float) -- Upper bound for the random contrast factor.
    """
    tf_image = tf.image.random_contrast(image, lower=lower, upper=upper)
    return tf_image, label


def random_hue(image, label: tuple, max_delta=0.5):
    """
    Data augmentation: Adjust the hue of RGB images by a random factor.

    Arguments:
        + image (tf.Tensor) -- Image to transform.
        + label (tf.Tensor) -- Label from the image.
        + max_delta (float) -- Maximum value for the random delta.
    """
    return tf.image.random_hue(image, max_delta=max_delta), label


def random_jpeg_quality(image, label: tuple, min_jpeg_quality=5, max_jpeg_quality=25):
    """
    Data augmentation: Randomly changes jpeg encoding quality for inducing jpeg noise.
    
    Arguments:
        + image (tf.Tensor) -- Image to transform.
        + label (tf.Tensor) -- Label from the image.
        + min_jpeg_quality () -- Minimum jpeg encoding quality to use.
        + max_jpeg_quality () -- Maximum jpeg encoding quality to use.
    """
    return tf.image.random_jpeg_quality(image, min_jpeg_quality=min_jpeg_quality, 
        max_jpeg_quality=max_jpeg_quality), label


def random_rotation(image, label: tuple, angle=45):
    """
    Data augmentation: Performs a random rotation of a image.

    Arguments:
        + image (tf.Tensor) -- Image to transform.
        + label (tf.Tensor) -- Label from the image.
        + angle (int) -- Rotation range, in degrees.
    """
    return tf.py_function(partial(_random_rotation, angle=angle), [image], tf.float32), label


def _random_rotation(image, angle=45):
    """
    Keras transformation: Performs a random rotation of a image.

    Arguments:
        + image (tf.Tensor) -- Image to transform.
        + angle (int) -- Rotation range, in degrees.
    """
    return tf.keras.preprocessing.image.random_rotation(image.numpy(), angle, 0, 1, 2)


def random_saturation(image, label: tuple, lower=0, upper=5):
    """
    Data augmentation: Adjust the saturation of RGB images by a random factor.
    
    Arguments:
        + image (tf.Tensor) -- Image to transform.
        + label (tf.Tensor) -- Label from the image.
        + lower (float) -- Lower bound for the random saturation factor.
        + upper (float) -- Upper bound for the random saturation factor.
    """
    return tf.image.random_saturation(image, lower=lower, upper=upper), label


def random_shear(image, label: tuple, intensity=45):
    """
    Data augmentation: Performs a random spatial shear of a image.
    
    Arguments:
        + image (tf.Tensor) -- Image to transform.
        + label (tf.Tensor) -- Label from the image.
        + intensity (int) -- Transformation intensity in degrees.
    """
    return tf.py_function(partial(_random_shear, intensity=intensity), [image], tf.float32), label


def _random_shear(image, intensity=45):
    """
    Keras transformation: Performs a random spatial shear of a image.
    
    Arguments:
        + image (tf.Tensor) -- Image to transform.
        + intensity (int) -- Transformation intensity in degrees.
    """
    return tf.keras.preprocessing.image.random_shear(image.numpy(), intensity, 0, 1, 2)
    

def random_zoom(image, label: tuple, zoom_range=(0.6, 0.6)):
    """
    Data augmentation: Performs a random spatial zoom of a image.

    Arguments:
        + image (tf.Tensor) -- Image to transform.
        + label (tf.Tensor) -- Label from the image.
        + zoom_range (Tuple of floats) -- Zoom range for width and height.
    """
    return tf.py_function(partial(_random_zoom, zoom_range=zoom_range), [image], tf.float32), label


def _random_zoom(image, zoom_range=(0.6, 0.6)):
    """
    Keras transformation: Performs a random spatial zoom of a image.
    
    Arguments:
        + image (tf.Tensor) -- Image to transform.
        + zoom_range (Tuple of floats) -- Zoom range for width and height.
    """
    return tf.keras.preprocessing.image.random_zoom(image.numpy(), zoom_range, 0, 1, 2)


"""
Code from opencv dnn module: 
    https://github.com/opencv/opencv/tree/master/samples/dnn/face_detector
"""

from __future__ import print_function
import os
import hashlib
import time
import sys
import xml.etree.ElementTree as ET
if sys.version_info[0] < 3:
    from urllib2 import urlopen
else:
    from urllib.request import urlopen


class _HashMismatchException(Exception):
    def __init__(self, expected, actual):
        Exception.__init__(self)
        self.expected = expected
        self.actual = actual
    def __str__(self):
        return 'Hash mismatch: {} vs {}'.format(self.expected, self.actual)


class _MetalinkDownloader(object):
    BUFSIZE = 10*1024*1024
    NS = {'ml': 'urn:ietf:params:xml:ns:metalink'}
    tick = 0

    def download(self, metalink_file):
        status = True
        for file_elem in ET.parse(metalink_file).getroot().findall('ml:file', self.NS):
            url = file_elem.find('ml:url', self.NS).text
            fname = file_elem.attrib['name']
            hash_sum = file_elem.find('ml:hash', self.NS).text
            print('*** {}'.format(fname))
            try:
                self.verify(hash_sum, fname)
            except Exception as ex:
                print('  {}'.format(ex))
                try:
                    print('  {}'.format(url))
                    with open(fname, 'wb') as file_stream:
                        self.buffered_read(urlopen(url), file_stream.write)
                    self.verify(hash_sum, fname)
                except Exception as ex:
                    print('  {}'.format(ex))
                    print('  FAILURE')
                    status = False
                    continue
            print('  SUCCESS')
        return status

    def print_progress(self, msg, timeout = 0):
        if time.time() - self.tick > timeout:
            print(msg, end='')
            sys.stdout.flush()
            self.tick = time.time()

    def buffered_read(self, in_stream, processing):
        self.print_progress('  >')
        while True:
            buf = in_stream.read(self.BUFSIZE)
            if not buf:
                break
            processing(buf)
            self.print_progress('>', 5)
        print(' done')

    def verify(self, hash_sum, fname):
        sha = hashlib.sha1()
        with open(fname, 'rb') as file_stream:
            self.buffered_read(file_stream, sha.update)
        if hash_sum != sha.hexdigest():
            raise _HashMismatchException(hash_sum, sha.hexdigest())


def maybe_download_model():
    """
    Download OpenCV Face Dectector model into designated model_path.

    Arguments:
        + model_path (str) -- model_path directory to OpenCV Caffe model.
    
    Files:
        + deploy.prototxt
        + res10_300x300_ssd_iter_140000_fp16.caffemodel
    """

    # Files path
    model_path = os.path.join(os.environ['HOME'], ".keras/opencv/face_detector/")
    model_caffe_path = os.path.join(model_path, "res10_300x300_ssd_iter_140000_fp16.caffemodel")
    proto_caffe_path = os.path.join(model_path, "deploy.prototxt")
    
    # Download model if it does not exists
    if not os.path.isfile(model_caffe_path) or not os.path.isfile(proto_caffe_path):
        if not os.path.exists(model_path):
            os.makedirs(model_path)

        # OpenCV files url
        url_files = {
            "weights.meta4":"https://raw.githubusercontent.com/opencv/opencv/master/samples/dnn/face_detector/weights.meta4",
            "deploy.prototxt":"https://raw.githubusercontent.com/opencv/opencv/master/samples/dnn/face_detector/deploy.prototxt",
        }

        # Download OpenCV script do download Face Detector weights
        for file in url_files:
            with open(os.path.join(model_path, file), "wb") as f:
                f.write(urlopen(url_files[file]).read())
            
        # Execute OpenCV script
        current_model_path = os.getcwd()
        os.chdir(model_path)
        _MetalinkDownloader().download('weights.meta4')
        os.chdir(current_model_path)


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import warnings
from tensorflow.keras.layers import *
from tensorflow.keras.models import Model
from tensorflow.keras.utils import get_file
from tensorflow.keras.backend import image_data_format

from .model_utils import _obtain_input_shape


# Download paths
VGGFACE_CE_L1_WEIGHTS_PATH = dict(
    name = "deep_expressions_tf_vggface_ce_l1_150.h5",
    path = ("https://github.com/deepexpressions/models/blob/master/"
            "emotions/deep_expressions_tf_vggface_ce_l1_150.h5?raw=true"),
    cache_subdir = "models/deep_expressions",
)
VGGFACE_CE_L3_WEIGHTS_PATH = dict(
    name = "deep_expressions_tf_vggface_ce_l3_150.h5",
    path = ("https://github.com/deepexpressions/models/blob/master/"
            "emotions/deep_expressions_tf_vggface_ce_l3_150.h5?raw=true"),
    cache_subdir = "models/deep_expressions",
)

VGGFACE_CK_L1_WEIGHTS_PATH = dict(
    name = "deep_expressions_tf_vggface_ck_l1_150.h5",
    path = ("https://github.com/deepexpressions/models/blob/master/"
            "emotions/deep_expressions_tf_vggface_ck_l1_150.h5?raw=true"),
    cache_subdir = "models/deep_expressions",
)
VGGFACE_CK_L3_WEIGHTS_PATH = dict(
    name = "deep_expressions_tf_vggface_ck_l3_150.h5",
    path = ("https://github.com/deepexpressions/models/blob/master/"
            "emotions/deep_expressions_tf_vggface_ck_l3_150.h5?raw=true"),
    cache_subdir = "models/deep_expressions",
)

VGGFACE_KDEF_L1_WEIGHTS_PATH = dict(
    name = "deep_expressions_tf_vggface_kdef_l1_150.h5",
    path = ("https://github.com/deepexpressions/models/blob/master/"
            "emotions/deep_expressions_tf_vggface_kdef_l1_150.h5?raw=true"),
    cache_subdir = "models/deep_expressions",
)
VGGFACE_KDEF_L3_WEIGHTS_PATH = dict(
    name = "deep_expressions_tf_vggface_kdef_l3_150.h5",
    path = ("https://github.com/deepexpressions/models/blob/master/"
            "emotions/deep_expressions_tf_vggface_kdef_l3_150.h5?raw=true"),
    cache_subdir = "models/deep_expressions",
)

VGGFACE_IFFG_L3_WEIGHTS_PATH = dict(
    name = "deep_expressions_tf_vggface_iffg_l3_150.h5",
    path = ("https://github.com/deepexpressions/models/blob/master/"
            "gestures/deep_expressions_tf_vggface_iffg_l3_150.h5?raw=true"),
    cache_subdir = "models/deep_expressions",
)


def _get_weights_path(weights, num_layers):
    """Return weights_path_dict based on weights name."""

    if weights == "ce" and num_layers == 1:
        return VGGFACE_CE_L1_WEIGHTS_PATH
    elif weights == "ce" and num_layers == 3:
        return VGGFACE_CE_L3_WEIGHTS_PATH

    elif weights in {"ck", "ck+"} and num_layers == 1:
        return VGGFACE_CK_L1_WEIGHTS_PATH
    elif weights in {"ck", "ck+"} and num_layers == 3:
        return VGGFACE_CK_L1_WEIGHTS_PATH

    if weights == "kdef" and num_layers == 1:
        return VGGFACE_KDEF_L1_WEIGHTS_PATH
    elif weights == "kdef" and num_layers == 3:
        return VGGFACE_KDEF_L1_WEIGHTS_PATH

    elif weights == "iffg" and num_layers == 3:
        return VGGFACE_IFFG_L3_WEIGHTS_PATH

    else:
        raise Exception("Couldn't find the model to download.")


def VGGFACE(weights=None, num_layers=1, num_classes=7, input_shape=(150, 150, 3)):
    """
    Load a FER Model trained on top of a VGGFACE implementation.

    Arguments:
        + weights -- Name of the database used to train the model. For details, 
        look at https://github.com/deepexpressions/models.
        + num_layers -- Number of layers in the neural network. Only 1 and 3 are available.
        + num_classes -- Number of output classes.
        + input_shape -- Input shape of the entrance image.

    Returns:
        + A model based on VGGFACE.
    """


    if not (weights in {"ce", "ck", "ck+", "kdef", "iffg", None} or os.path.exists(weights)):
        raise ValueError("The `weights` argument should be either "
                         "`None` (random initialization), `ce`, `ck`, `ck+`, `iffg`"
                         "or the path to the weights file to be loaded.")

    if weights in {"ce", "ck", "ck+"} and num_classes != 7:
        raise ValueError("If using `weights` as [`ce`, `ck`, `ck+`], "
                         "`num_classes` should be 7.")

    if weights == "iffg" and num_classes != 9:
        raise ValueError("If using `weights` as `iffg`, `num_classes` should be 9.")

    if weights in {"ce", "ck", "ck+", "kdef"} and input_shape != (150, 150, 3):
        warnings.warn("This model was trained with `input_shape = (150, 150, 3)`")

    if weights == "iffg" and input_shape != (224, 224, 3):
        warnings.warn("This model was trained with `input_shape = (224, 224, 3)`")


    # Determine model name
    model_name = "vggface" if weights is None else "vggface_" + weights


    # Determine proper input shape
    input_shape = _obtain_input_shape(input_shape,
                                      default_size=150,
                                      min_size=48,
                                      data_format=image_data_format(),
                                      require_flatten=True,
                                      weights=weights)

    # Model input
    inputs = Input(shape=input_shape)

    # Block 1
    x = Conv2D(64, (3, 3), activation="relu", padding="same", name="block1_conv1")(inputs)
    x = Conv2D(64, (3, 3), activation="relu", padding="same", name="block1_conv2")(x)
    x = MaxPooling2D((2, 2), strides=(2, 2), name="block1_pool")(x)

    # Block 2
    x = Conv2D(128, (3, 3), activation="relu", padding="same", name="block2_conv1")(x)
    x = Conv2D(128, (3, 3), activation="relu", padding="same", name="block2_conv2")(x)
    x = MaxPooling2D((2, 2), strides=(2, 2), name="block2_pool")(x)

    # Block 3
    x = Conv2D(256, (3, 3), activation="relu", padding="same", name="block3_conv1")(x)
    x = Conv2D(256, (3, 3), activation="relu", padding="same", name="block3_conv2")(x)
    x = Conv2D(256, (3, 3), activation="relu", padding="same", name="block3_conv3")(x)
    x = MaxPooling2D((2, 2), strides=(2, 2), name="block3_pool")(x)

    # Block 4
    x = Conv2D(512, (3, 3), activation="relu", padding="same", name="block4_conv1")(x)
    x = Conv2D(512, (3, 3), activation="relu", padding="same", name="block4_conv2")(x)
    x = Conv2D(512, (3, 3), activation="relu", padding="same", name="block4_conv3")(x)
    x = MaxPooling2D((2, 2), strides=(2, 2), name="block4_pool")(x)

    if num_layers == 1:
        # Classification Block
        x = MaxPooling2D((2, 2), strides=(2, 2), name="extra_pool")(x)
        x = Flatten(name="flatten")(x)
        x = Dropout(0.20, name="flatten/drop1")(x)
        x = Dense(7, activation="softmax", name="predictions")(x)
    
    elif num_layers == 3:
        # Classification Block
        x = Flatten(name="flatten")(x)
        x = Dense(256, activation="relu", name="fc1")(x)
        x = Dropout(0.3, name="fc1/drop1")(x)
        x = Dense(128, activation="relu", name="fc2")(x)
        x = Dropout(0.2, name="fc2/drop1")(x)
        x = Dense(64, activation="relu", name="fc3")(x)
        x = Dense(num_classes, activation="softmax", name="predictions")(x)

    else:
        raise ValueError("`num_layers` must be 1 or 3.")

    # Create model
    model = Model(inputs, x, name=model_name)

    # Load weights
    if weights in {"ce", "ck", "ck+", "kdef", "iffg"}:
        wpd = _get_weights_path(weights, num_layers)
        weights_path = get_file(wpd["name"], wpd["path"], cache_subdir=wpd["cache_subdir"])
        model.load_weights(weights_path)

    elif weights is not None:
        model.load_weights(weights)

    return model